####################################################
# ClampConnect Utility                             #
# Author: LambdaFox                                #
# Date: 10 August 2024                             #
# Version: 0.1.0 (initial test build)              #
# License: Clear attribution & link to website     #
# Website: http://LambdaFox.com/clamp-connect      #
# Proof of concept for interfacing with Fluke 376  #
# connect devices over Bluetooth using QBluetooth  #
####################################################
# Dependencies:                                    #
# pip install pyqt6                                #
####################################################
# No warranty whether expressed or implied. User   #
# assumes all risk.                                #
####################################################
import sys
from PyQt6.QtCore import QByteArray
from PyQt6 import QtWidgets, uic
from PyQt6.QtBluetooth import(
    QBluetoothDeviceDiscoveryAgent,
    QLowEnergyController,
    QLowEnergyService,
    QBluetoothUuid)

class MainWindow(QtWidgets.QMainWindow):
    """Main Qt application"""
    TARGET_DEVICE_NAME          = "376FC"                                                   #Bluetooth device name
    TARGET_SERVICE_UUID         = QBluetoothUuid("b6981800-7562-11e2-b50d-00163e46f8fe")    #Service UUID
    TARGET_CHARACTERISTIC_UUID  = QBluetoothUuid("b6982901-7562-11e2-b50d-00163e46f8fe")    #Characteristic UUID

    def __init__(self):
        super().__init__()
        uic.loadUi('user_interface.ui', self)
        self.pushButton.clicked.connect(self.start_scan)

        self.deviceDiscoveryAgent = QBluetoothDeviceDiscoveryAgent(self)
        self.deviceDiscoveryAgent.deviceDiscovered.connect(self.deviceDiscovered)
        self.deviceDiscoveryAgent.setLowEnergyDiscoveryTimeout(15000)
        self.deviceDiscoveryAgent.errorOccurred.connect(self.scan_error)
        self.deviceDiscoveryAgent.finished.connect(self.scan_finished)
        self.deviceDiscoveryAgent.canceled.connect(self.scan_finished)

        self.m_control = None
        self.m_service = None

    def deviceDiscovered(self, device):
        self.log(1, f"Discovered Device: {device.name()} ({device.address().toString()})")
        if device.name() == self.TARGET_DEVICE_NAME:
            self.log(1, f"Found target device: {self.TARGET_DEVICE_NAME}")
            self.m_control = QLowEnergyController.createCentral(device, self)
            self.m_control.connected.connect(self.onConnected)
            self.m_control.disconnected.connect(self.onDisconnected)
            self.m_control.errorOccurred.connect(self.onError)
            self.m_control.serviceDiscovered.connect(self.serviceDiscovered)
            self.m_control.discoveryFinished.connect(self.serviceScanDone)
            self.m_control.connectToDevice()

    def onConnected(self):
        self.log(1, "Device connected, discovering services...")
        self.m_control.discoverServices()

    def onDisconnected(self):
        self.log(1, "Device disconnected")
        self.m_control = None
        self.m_service = None

    def onError(self, error):
        self.log(1, f"Controller Error: {error}")

    def serviceDiscovered(self, service_uuid):
        self.log(1, f"Service discovered: {service_uuid.toString()}")
        if service_uuid == self.TARGET_SERVICE_UUID:
            self.log(1, "Found target service")
            self.m_service = self.m_control.createServiceObject(service_uuid, self)
            if self.m_service:
                self.m_service.stateChanged.connect(self.serviceStateChanged)
                self.m_service.characteristicChanged.connect(self.characteristicChanged)
                self.m_service.descriptorWritten.connect(self.descriptorWritten)
                self.m_service.discoverDetails()
            else:
                self.log(1, "Failed to create service object")

    def serviceScanDone(self):
        self.log(1, "Service scan completed")
        if not self.m_service:
            self.log(1, "Target service not found.")

    def serviceStateChanged(self, state):
        self.log(1, f"Service state changed: {state}")
        if state == QLowEnergyService.ServiceState.ServiceDiscovered:
            self.log(1, "Service discovered. Listing characteristics...")
            self.list_characteristics()
            self.setup_notifications()
        else:
            self.log(1, f"Service state: {state}")

    def list_characteristics(self):
        characteristics = self.m_service.characteristics()
        if not characteristics:
            self.log(1, "No characteristics found in the service.")
            return
        for char in characteristics:
            self.log(1, f"Characteristic UUID: {char.uuid().toString()}")
            self.log(1, f"   Properties: {char.properties()}")

    def setup_notifications(self):
        try:
            characteristic = self.m_service.characteristic(self.TARGET_CHARACTERISTIC_UUID)
            if not characteristic.isValid():
                self.log(1, "Target characteristic not found.")
                return
    
            self.log(1, "Setting up notifications for the characteristic...")
    
            # Create a QBluetoothUuid using DescriptorType
            cccd_uuid = QBluetoothUuid(QBluetoothUuid.DescriptorType.ClientCharacteristicConfiguration)
            descriptor = characteristic.descriptor(cccd_uuid)
            if not descriptor.isValid():
                self.log(1, "CCCD Descriptor not found. Cannot subscribe to notifications.")
                return
    
            self.m_service.writeDescriptor(descriptor, QByteArray(b"\x01\x00"))
            self.log(1, "Notification subscription request sent.")
    
        except Exception as e:
            self.log(1, f"Error during setup_notifications: {e}")


    def characteristicChanged(self, characteristic, value):
        self.log(1, f"Characteristic {characteristic.uuid().toString()} changed. New value: {value.data().decode('utf-8')}")
        self.lineEdit.setText(value.data().decode('utf-8'))

    def descriptorWritten(self, descriptor, value):
        self.log(1, f"Descriptor {descriptor.uuid().toString()} written. Value: {value.toHex().data().decode()}")

    def log(self, level, message):
        if level == 1:
            print(f"[Level {level}]: {message}")

    def scan_error(self, error):
        self.log(1, f"Scan Error: {error}")

    def scan_finished(self):
        self.log(1, "Device scan finished.")

    def start_scan(self):
        self.log(1, "Starting device scan...")
        self.deviceDiscoveryAgent.start(QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.LowEnergyMethod)

def main():
    try:
        app = QtWidgets.QApplication(sys.argv)
        window = MainWindow()
        window.show()
        sys.exit(app.exec())
    except Exception as e:
        print(f"Error in main: {e}")

if __name__ == '__main__':
    main()
